#!/usr/bin/env python
"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""

from __future__ import print_function

import argparse
from datetime import datetime
import math

import ddlrun.utilities as utilities
from ddlrun.error import eprint

def get_gpu_per_socket(nvidia_topo_results, threads, cores):
    """Uses nvidia-smi topo -m output to determine GPUs per socket"""
    #Get rid of the header before processing starts
    nvidia_topo_results.pop(0)

    #affinity_dict will be of format:
    #affinity_dict[socket_num] = gpu count for that socket
    affinity_dict = {}
    format_recognized = True
    for line in nvidia_topo_results:
        line_elements = line.split()
        #Purpose of the for loop is build affinity_dict based on
        #the number of accelerators in the system as detected by
        #nvidia-smi command.
        if line_elements and "GPU" in line_elements[0]:
            #Last column in each line is CPU Affinity. Splitting it
            #to get range of CPUs associated with this GPU.
            try:
                core_elements = line_elements[-1].split('-')

                #Nvidia-smi spits out CPU affinity, but we want cores.
                #num of cpus by nvidia-smi = # of cores * # of threads per core
                socket_num = int(core_elements[1]) // (threads * cores)
                if socket_num in affinity_dict:
                    affinity_dict[socket_num] += 1
                else:
                    affinity_dict[socket_num] = 1

            except (ValueError, IndexError):
                eprint("WARNING: nvidia-smi topo -m output not recognized. "\
                            "Defaulting to 2 GPUs per socket")
                format_recognized = False
                break

    gpus_per_socket = -1
    if not format_recognized:
        gpus_per_socket = 2
    elif affinity_dict:
        gpus_per_socket = affinity_dict[list(affinity_dict.keys())[0]]

    return gpus_per_socket


def read_args():  #pylint: disable=too-many-branches,too-many-statements
    """Handles the set up of CLI parameters, and errors that may occur from invalid parameters"""
    #~ set up of the arguments
    parser = argparse.ArgumentParser(
        description='Create rank file for specified hosts.')
    parser.add_argument(
        '--aisles',
        dest='aisles',
        default=1,
        type=int,
        help='Number of aisles. If no option given 1 is assumed.')
    parser.add_argument(
        '-r',
        '--racks',
        dest='racks',
        default=1,
        type=int,
        help='Number of racks in each aisle. If no option given 1 is assumed.')
    parser.add_argument(
        '-n',
        '--nodes',
        dest='nodes',
        default=-1,
        type=int,
        help=
        'Number of nodes per rack. If no option given the number of hosts given is assumed.'
    )
    parser.add_argument(
        '-a',
        '--accelerators',
        dest='accelerators',
        default=-1,
        type=int,
        help='Number of accelerators per node. If no option given the number in '
        + 'the current machine is assumed. Error on values larger than 16.')
    parser.add_argument(
        '-s',
        '--sockets',
        dest='sockets',
        default=-1,
        type=int,
        help=
        'Number of CPU sockets per node. Values can be 1 or any even number 2 '
        +
        'or larger. If no option given the number in the current machine is assumed.'
    )
    parser.add_argument(
        '-c',
        '--cores',
        dest='cores',
        default=-1,
        type=int,
        help='Number of cores per socket. If no option given'
        ' the number in the current machine is assumed.')
    parser.add_argument(
        '-g',
        '--gpu_per_socket',
        dest='gpu_per_socket',
        default=-1,
        type=int,
        help='Number of GPUs per socket. If no option given'
        ' the number is calculated based on current machine\'s configuration.')
    parser.add_argument(
        "--all_cores",
        action="store_false",
        dest="split",
        help=
        "Each instance can use all cores. Default is to equally divide cores between instances."
    )
    parser.add_argument(
        "--machine_report",
        action="store_true",
        default=False,
        dest="machine_report",
        help=argparse.SUPPRESS)
    parser.add_argument(
        'hosts_cli', metavar='hosts', help='Comma seperated list of hosts.')
    args = parser.parse_args()

    #~ Transforming hosts CSV into a list
    args.hosts = args.hosts_cli.split(",")

    #~ If no value was passed for nodes, set it to the number of hosts
    if args.nodes == -1:
        args.nodes = (len(args.hosts) // args.racks) // args.aisles
    else:
        args.nodes = args.nodes

    #~ If no value was passed for sockets or cores get the values from lscpu instead
    if (args.sockets == -1) or (args.cores == -1) or (
            args.gpu_per_socket == -1):
        try:
            _, lscpu_results, _ = utilities.run_command_capture("lscpu")
            lscpu_results = lscpu_results.splitlines()
        except ValueError:
            parser.error(
                'lscpu failed to run. Please provide values for sockets and cores.'
            )
        except OSError:
            parser.error(
                'lscpu failed to run. Please provide values for sockets and cores.'
            )

    if args.sockets == -1:
        try:
            args.sockets = int(''.join(
                filter(
                    str.isdigit,
                    next((s for s in lscpu_results if "Socket(s):" in s),
                         None))))
        except TypeError:
            parser.error(
                'Unable to parse lscpu. Please provide values for sockets and cores.'
            )

    if args.cores == -1:
        try:
            args.cores = int(''.join(
                filter(
                    str.isdigit,
                    next(
                        (s
                         for s in lscpu_results if "Core(s) per socket:" in s),
                        None))))
        except TypeError:
            parser.error(
                'Unable to parse lscpu. Please provide values for sockets and cores.'
            )

    #~ If no value was passed for accelerators get the values from nvidia-smi instead
    if args.accelerators == -1:
        try:
            nvidia_smi_retcode, nvidia_smi_results, _ = utilities.run_command_capture(
                "nvidia-smi --query-gpu=count --format=csv,noheader")
            if not nvidia_smi_retcode:
                parser.error(
                    'nvidia-smi failed to run. Please provide value for accelerators per node.'
                )
            args.accelerators = int(nvidia_smi_results.splitlines()[0])
        except ValueError:
            parser.error(
                'nvidia-smi failed to run. Please provide value for accelerators per node.'
            )
        except OSError:
            parser.error(
                'nvidia-smi failed to run. Please provide value for accelerators per node.'
            )

    #~ If no value was passed for gpu_per_socket get the values from nvidia-smi instead
    if args.gpu_per_socket == -1:
        try:
            _, nvidia_topo_results, _ = utilities.run_command_capture(
                "nvidia-smi topo -m")
            args.nvidia_topo_results = nvidia_topo_results.splitlines()
            args.threads = int(''.join(
                filter(
                    str.isdigit,
                    next(
                        (s
                         for s in lscpu_results if "Thread(s) per core:" in s),
                        None))))
        except (ValueError, OSError):
            parser.error(
                'nvidia-smi failed to run. Please provide value for GPU per socket'
            )
        except TypeError:
            parser.error(
                'Unable to parse lscpu. Please provide values for threads per core'
            )

        args.gpu_per_socket = get_gpu_per_socket(args.nvidia_topo_results,
                                                 args.threads, args.cores)

    #~ Invalid usage checks.
    #~ aisles, racks, nodes, accelerators, cores, sockets all need to be above zero.
    if args.aisles < 1:
        parser.error('aisles need to be greater than 0.')
    if args.racks < 1:
        parser.error('Racks need to be greater than 0.')
    if args.nodes < 1:
        parser.error('Nodes need to be greater than 0.')
    if args.accelerators < 1:
        parser.error('Accelerators need to be greater than 0.')
    if args.cores < 1:
        parser.error('Cores need to be greater than 0.')
    if args.sockets < 1:
        parser.error('Sockets needs to be greater than 0.')
    if args.gpu_per_socket < 1:
        parser.error('GPUs per socket needs to be greater than 0.')
    if args.sockets * args.cores < args.accelerators:
        parser.error(
            'Invalid argument combination. The number of cores should ' +
            ' be greater or equal to the number of accelerators in each machine.'
        )

    if args.accelerators < args.gpu_per_socket:
        args.gpu_per_socket = args.accelerators

    #~ The aisles and racks should have equal number of nodes.
    node_count = args.aisles * args.racks * args.nodes
    if (node_count != len(args.hosts)) or (len(args.hosts) % args.racks !=
                                           0) or (len(args.hosts) % args.aisles
                                                  != 0):
        parser.error('Invalid argument combination.' +
                     ' The number of aisles: ' + str(args.aisles) +
                     ', racks per aisle: ' + str(args.racks) +
                     ', and nodes per rack: ' + str(args.nodes) +
                     ' does not match the count of hosts given: ' +
                     str(len(args.hosts)) + '.')

    return args


def product(input_list):
    """Computes the product of a given list."""
    retval = 1
    for val in input_list:
        retval *= val
    return retval


def set_val(indexes, counts, input_list, value):
    """Given <input_list> of size product(<counts>), which represents a multidimensional matrix
    whose dimensions are represented by <counts>, this function sets the value at the location
    represented by <indexes> to <value>.
    """
    index = 0
    for depth in range(len(counts)):
        index += indexes[depth] * product(counts[depth + 1::])
    input_list[index] = value
    return input_list


def get_val(indexes, counts, input_list):
    """Given <input_list> of size product(<counts>), which represents a multidimensional matrix
    whose dimensions are represented by <counts>, this function returns the value at the location
    represented by <indexes>.
    """
    index = 0
    for depth in range(len(counts)):
        index += indexes[depth] * product(counts[depth + 1::])
    return input_list[index]


def rankify(counts, rank=0, depth=0, indexes=None, ranks=None):
    """This function creates a list of column major indexes for a
    matrix with the dimensions specified by counts.
    """
    indexes = [] if indexes is None else indexes
    ranks = [] if ranks is None else ranks

    if not ranks:  #~ Initialize the return list.
        ranks = [0] * product(counts)
    if depth < len(counts):
        for i in range(counts[len(counts) - depth - 1]):
            ranks, rank = rankify(counts, rank, depth + 1, [i] + indexes,
                                  ranks)
    else:
        set_val(indexes, counts, ranks, rank)
        rank += 1
    return ranks, rank


def rankfile_header(counts, cores, sockets, hosts):
    """Generate the first lines for the rankfile."""
    print("#%s by rank_gen" % datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    print("#host = %s" % ",".join(hosts))
    print("#aisles = %s" % counts[0])
    print("#racks = %s" % counts[1])
    print("#nodes = %s" % counts[2])
    print("#accelerators = %s" % counts[3])
    print("#sockets = %s" % sockets)
    print("#cores = %s" % cores)
    print("")


def rankfile_body(counts, cores, sockets, hosts, split, ranks, node_dim, gpu_per_socket):  #pylint: disable=too-many-arguments,too-many-locals
    """The node_dim argument specifies the last dimension that is outside of the node.
    If the number of cores isn't evenly divisible by the accelerators per socket,
    the last rank on the socket gets the extra cores.
    """
    total_instances = product(counts)
    instances_per_socket = max(
        math.ceil((total_instances / len(hosts)) / (1.0 * sockets)),
        gpu_per_socket)
    if split:
        cores_per_instance = cores // instances_per_socket
    else:
        cores_per_instance = cores

    socket = instance_this_socket = current_host = 0

    for rank_index, _ in enumerate(ranks):
        if not split:
            cores_begin = 0
        else:
            cores_begin = instance_this_socket * cores_per_instance

        #~ If not splitting cores, or if on the last rank of the socket, cores_end is the last core.
        if not split or instance_this_socket >= instances_per_socket - 1:
            cores_end = cores - 1
        else:
            cores_end = (instance_this_socket + 1) * cores_per_instance - 1

        print("rank %d=%s\t   slot=%d:%d-%d" %
              (ranks[rank_index], hosts[current_host], socket, cores_begin,
               cores_end))
        instance_this_socket += 1

        #~ If we've finished this socket, move on to the next socket.
        if instance_this_socket >= instances_per_socket:
            socket += 1
            instance_this_socket = 0

        #~ If the next rank index is evenly divisible by the total instances within a node,
        #~ then we know that it is the start of the next host.
        if socket >= sockets or (rank_index + 1) % product(
                counts[node_dim + 1::]) == 0:
            current_host += 1
            socket = 0
            instance_this_socket = 0
            print("")


def main():
    """Program entry point."""
    args = read_args()

    if args.machine_report:
        print("%d,%d,%d" % (args.accelerators, args.cores, args.sockets))
    else:
        counts = [args.aisles, args.racks, args.nodes, args.accelerators]
        ranks, _ = rankify(counts)
        rankfile_header(counts, args.cores, args.sockets, args.hosts)
        rankfile_body(counts, args.cores, args.sockets, args.hosts, args.split,
                      ranks, 2, args.gpu_per_socket)


if __name__ == "__main__":
    main()
