"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""
import logging
import subprocess
import os
import socket
import psutil

# These ssh options force the usage of private/public keys and don't prompt the user for input.
SSH_ARGS = {
    "-oPasswordAuthentication=no", "-oStrictHostKeyChecking=no",
    "-oBatchMode=yes"
}

def run_command_capture(cmd):
    """Run a shell command and capture the ret_code, stdout and stderr."""
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        universal_newlines=True)
    output, error = process.communicate()

    return process.returncode == 0, output, error


def run_command(cmd):
    """Run a shell command and don't collect the output."""
    process = subprocess.Popen(cmd, shell=True, universal_newlines=True)
    process.wait()
    return process.returncode

def run_program_with_ssh(cmd, args):
    """Run each instance of the program with a SSH call."""
    return_code = 0
    processes = []
    num_accelerators = int(args.topology.split("x")[0])
    group_id = get_group_id()
    index = 0
    working_dir = os.getcwd()
    for host in args.hosts:
        local_host = NetUtils.is_local_host(host)
        for gpu in range(num_accelerators):
            ddl_options = get_flex_ddl_options(args.num_procs, index, gpu, group_id, args.ddl_args)
            cmd_to_run = "%s %s bash -c '%s'" % \
                         (ddl_options, make_arg_list(args.env_vars), cmd)
            if not local_host:
                cmd_to_run = "ssh %s %s \"cd %s > /dev/null 2>&1; %s\"" % \
                             (make_arg_list(SSH_ARGS), host, working_dir, cmd_to_run)
            logging.getLogger("ddlrun").critical("+ %s", cmd_to_run)
            process = subprocess.Popen(cmd_to_run, shell=True, universal_newlines=True)
            processes.append(process)
            index += 1
    for process in processes:
        process.wait()
        # Return only the first non-zero return code encountered
        if return_code == 0:
            return_code = process.returncode
    return return_code

def append_all_text(filename, text):
    """Appends to a text file, creating it if it doesn't exist."""
    with open(filename, "a+") as output_file:
        output_file.write(text)


def read_all_text(filename):
    """Reads all test from a text file."""
    with open(filename, 'r') as text_file:
        text = text_file.read()
    return text


def make_arg_list(arglist):
    """Converts a list of strings to a properly quoted arg list."""
    return ' '.join(str(arg) for arg in arglist)


def make_prog_arg_list(arglist):
    """Converts a list of strings to a properly quoted arg list."""
    quotable_chars = set(" !\"#$&()*,;<>?^`{|}")
    return ' '.join('"' + arg + '"' if any((c in quotable_chars)
                                           for c in arg) else arg
                    for arg in arglist)


def get_flex_ddl_options(count, index, dev_index, group_id, ddl_args):
    """Returns DDL options to use in flex mode."""
    return "DDL_OPTIONS='-ddl_worker_count %s -ddl_worker_index %s \
           -ddl_device_index %s -ddl_group_id %s %s'" % \
           (count, index, dev_index, group_id, make_arg_list(ddl_args))


def get_group_id():
    """Returns the group identifier"""
    import pyddl
    return pyddl.generate_group_id()


def make_temp_rankfile(host, name, working_dir):
    """Makes a simple temporary rankfile.

    Args:
        host: The single host to put in the rankfile.
        name: The suffix to use for the rankfile.
        working_dir: The temporary location to store the rankfile.

    Returns:
        The full rankfile path.
    """
    temp_rank_filename = os.path.join(working_dir, "temp_rankfile_%s" % (name))
    with open(temp_rank_filename, "w") as temp_rank_file:
        temp_rank_file.write("rank 0=%s slot=0" % (host))
    return temp_rank_filename


def mpirun_command(command,
                   host=None,
                   rank_file=None,
                   conda_env=None,
                   args=None):
    """Runs the given command using mpirun.

    If a host is given, then a temporary rankfile is made to run on that host.
    Otherwise the given rankfile is used, if provided.

    Args:
        command: The command to launch via mpirun.
        host: The host to generate a rankfile for.
        rank_file: Explicit rank_file to use.
        args: Parsed program arguments.

    Returns:
        The full mpirun command.
    """
    if rank_file is None and args is not None and host is not None:
        rank_file = make_temp_rankfile(host, host, args.working_dir)
    command_to_run = make_mpirun_command(command, args, rank_file, conda_env)
    return run_command_capture(command_to_run)


def make_mpirun_command(command,
                        args,
                        rank_file=None,
                        conda_env=None):
    """Constructs the string mpirun command.

    Args:
        command: The command to launch via mpirun.
        args: The arguments passed to ddlrun
        rank_file: The rankfile to use.
        conda_env: If passed in, activates the conda env before executing command.

    Returns:
        The full mpirun command.
    """
    result = args.mpirun_exe

    if args.debug_timeout is not None:
        result += " -timeout " + str(args.debug_timeout)

    if args.mpi_args is not None:
        result += " " + make_arg_list(args.mpi_args)

    result += " -n 1"

    if rank_file is not None:
        result += " --rankfile " + str(rank_file)

    result += " bash -c '"

    if conda_env is not None:
        result += "source %s && conda activate %s  > /dev/null 2>&1 && " % \
                (args.conda_setup_script, conda_env)

    result += str(command) + "'"

    return result


class NetUtils: # pylint: disable=too-few-public-methods
    """Network utilities"""
    _local_ips_cache = set()

    @staticmethod
    def is_local_host(hostname):
        """Returns True if hostname resolves to a local ip address"""
        if not NetUtils._local_ips_cache:
            NetUtils._local_ips_cache = {ip.address for _, info in psutil.net_if_addrs().items()
                                         for ip in info if ip.family == socket.AF_INET or
                                         ip.family == socket.AF_INET6}
        hostname_info = None
        try:
            hostname_info = socket.getaddrinfo(hostname, None, socket.AF_INET)
        except socket.gaierror:
            # Check if there is an IPV6 address associated to this hostname instead
            try:
                hostname_info = socket.getaddrinfo(hostname, None, socket.AF_INET6)
            except socket.gaierror:
                # Keep going if hostname could not be found, because we will test connectivity later
                pass
        if hostname_info:
            # structure
            #[(family, socktype, proto, canonname, sockaddr)...] with sockaddr being (address, ...)
            return hostname_info[0][4][0] in NetUtils._local_ips_cache
        return False
